
' ===========================================================================
' Soundausgabe auf der Open-Maxi an DA[1]
' In OCBASIC 1.09eee von Dietmar Harlos am 27. Juni 2012
' ===========================================================================

' Dieses Demoprogramm gibt eingesampelte Klnge an einem an DA[1] angeschlos-
' senen Piezo-Tner (ohne Elektronik) wieder. Die Soundausgabe findet im
' Hintergrund, whrend der normalen Programmabarbeitung statt.

' Bei 8 Bit Auflsung sind PWM-Frequenzen bis 78 kHz mglich. Allerdings
' zeigten Messungen, da es dabei (vermutlich durch den Piezo) zu Verzer-
' rungen kommt. Sinnvoller sind 39 kHz. In diesem Beispiel werden nur 16 kHz
' verwendet, da dadurch der beste Kompromi zwischen Klangqualitt und ein-
' facher Programmlogik erzielt wird. Durch Mittelung werden Zwischenwerte
' berechnet und dadurch die Klangqualitt verbessert.

' Mgliche Verbesserungen:

' - Durch die niedrige PWM-Frequenz von 16 kHz ist das Spiegelspektrum
'   zwischen 12 und 16 kHz gut hrbar. Es ist daher sinnvoll, einen RC-
'   Tiefpa mit 4 kHz Grenzfrequenz zu verwenden.

' - Am besten einen ohmschen, hochohmigen Verbraucher an DA[1] anschlieen.
'   Mit einem als Emitterfolger geschalteten BC548-Transistor und einem
'   12 Ohm Vorwiderstand habe ich einen 8 Ohm Lautsprecher angeschlossen
'   und gute Zimmerlautstrke erreicht.

' - Bei 32 kHz PWM-Frequenz knnte man mittels Faltung mit der Impulsantwort
'   eines Tiefpafilters die Spiegelspektren reduzieren. Es werden sozusagen
'   Zwischenwerte berechnet. Es mu nur jede vierte Multiplikation und
'   Addition ausgefhrt werden, wenn die "Zwischensamples" auf 0 gesetzt
'   werden. Ob die Faltung zuviel Zeit bentigt, habe ich nicht getestet.

' - Um eine PWM-Frequenz von 16 kHz zu erzielen, konnte das Timer-Counter-
'   Modulo-Register nicht auf den bei 8 Bit Auflsung blichen Wert 255
'   gesetzt werden, sondern auf 311. Dadurch ist die Tonausgabe um 1.72 dB
'   leiser. Wer mchte, kann entsprechend skalieren.

INCLUDE "omax.def"

DIM z,ende BYTE
DIM position WORD

DEFINE laenge 58422

' ---------------------------------------------------------------------------
' Soundausgabe auf der Open-Maxi an DA[1]. Die folgenden Subroutinen sollten
' sinnvollerweise in einer INCLUDE-Datei untergebracht werden.
' ---------------------------------------------------------------------------

' Das Schlsselwort INLASM kennzeichnet die folgende PROCEDURE als
' Assemblerroutine ("IAR"). Dadurch, und da sie unmittelbar am Anfang
' des Programms steht, wird sie zu einer IIAR, also zu einer Assembler-
' Interruptroutine, die bei einem aktivierten User-Interrupt gestartet wird.

PROC iiar_tpm2ofl INLASM
! cbeqa #iTPM2OFL,is_tpm2ofl
! jmp iiar_tpm2ofl_next+2   'zur naechsten IIAR springen

#is_tpm2ofl
! brset 0,ende,is_tpm2ofl_rts
! ldhx position
! lda soundtab,x
! inc z
! brset 1,z,is_tpm2ofl_is2
! add soundtab-1,x
! rora
! bra is_tpm2ofl_setv
#is_tpm2ofl_is2
! clr z
! aix #1
! sthx position
! cmphx #laenge
! bne is_tpm2ofl_setv
! mov #1,ende
#is_tpm2ofl_setv
! clr TPM2C0VH
! sta TPM2C0VL
#is_tpm2ofl_rts
! clc
! rts

END PROC

' Das TPM2-Modul fr Soundausgabe konfigurieren

' DA[1] liegt an TPM2CH0
' DA[2] liegt an TPM2CH1

PROC inittpm2 INLASM
! mov #%01001010,TPM2SC     'TPM clock input = 20E6/4 = 5 MHz & Timer Overflow Interrupt Enable
! ldhx #311                 'Timer 2 Counter Modulo Registers (TPMxMODH:TPMxMODL)
! sthx TPM2MODH             'von 0 bis inklusive 311 zhlen (also Modulo 312) - (kein 100% duty cycle)
! lda #%00101000            'PWM-Frequenz: 20E6/4/312 = 16.025 kHz
! sta TPM2C0SC              'Timer 2 Channel 0 Status and Control Register \ Edge-aligned PWM
! sta TPM2C1SC              'Timer 2 Channel 1 Status and Control Register /  High-true pulses
! rts
END PROC

#iiar_tpm2ofl_next

' ---------------------------------------------------------------------------
' Ende der IIAR. Hier koennen weitere IIARs folgen.
' ---------------------------------------------------------------------------

#main        'durch die IIAR am Programmbeginn wird ein "GOTO main" erzeugt

' ---------------------------------------------------------------------------
' Das Hauptprogramm
' ---------------------------------------------------------------------------

position=1
ende=0
z=0
UIRTpm2Ofl=ON
inittpm2

PrintSpc=ON

WHILE TRUE
  PRINT "Spiele Sound an DA[1]";position;ende
  PAUSE 10
  IF ende THEN
    PRINT "Beginne von neuem"
    PAUSE 75
    position=1
    ende=0
    z=0
  END IF
WEND

END

' ---------------------------------------------------------------------------
' Sounddaten im Format "8-bit Unsigned", 8 kHz Sample-Frequenz
' ---------------------------------------------------------------------------

' Aus "Per Anhalter ins All": Eddie, der Bordcomputer eines Raumschiffs, der
' alle Anwesenden mit seiner endlos guten Laune tyrannisiert.

TABLE soundtab BYTE "sound2.tab"

