
' ===========================================================================
' Soundausgabe auf der Open-Maxi an DA[1]
' In OCBASIC 1.09eee von Dietmar Harlos am 28. Juni 2012
' ===========================================================================

' Als Bonus hier noch ein aufwendigeres Verfahren, das zu einer besseren
' Klangqualitt fhrt. Es arbeitet mit 32 kHz PWM-Frequenz, um die 8-kHz-
' Samples abzuspielen. Um das zu bewerkstelligen, ist die einfachste Methode,
' einfach viermal dasselbe Sample abzuspielen. Das fhrt aber zu Spiegel-
' spektren und somit zu einer schlechten Klangqualitt. In diesem Programm
' wird beim bergang vom 8-kHz- in den 32-kHz-Bereich jedes Sample viermal
' wiederholt, was einer Faltung ("Konvolution") mit einer Rechteckfunktion
' entspricht, und dann der Mittelwert ber 4 Samples berechnet. Jetzteres
' entspricht einer weiteren Faltung mit einer Rechteckfunktion. Insgesamt
' betrachtet fhrt der Algorithmus eine Faltung mit einer Dreiecksfunktion
' durch. Die Impulsantwort ist also dreiecksfrmig. Der Mittelwert wird
' zum Schlu mit 9 Bit Auflsung ausgegeben.

' Mgliche Verbesserungen:

' - Eine weitere interessante Verbesserung wre, die Samplewerte bei der
'   Mittelung mit der Gau-Funktion ("Glockenkurve") zu wichten. Dadurch
'   ergibt sich eine Impulsantwort, die ebenfalls Glockenform hat. In
'   Simulationen hat sich die Berechnung "(38*sb0+89*sb1+89*sb2+38*sb3)/256"
'   als brauchbar erwiesen.

' - Um eine PWM-Frequenz von 32 kHz zu erzielen, konnte das Timer-Counter-
'   Modulo-Register nicht auf den bei 9 Bit Auflsung blichen Wert 511
'   gesetzt werden, sondern auf 623. Dadurch ist die Tonausgabe um 1.72 dB
'   leiser. Wer mchte, kann entsprechend skalieren.

INCLUDE "omax.def"

DIM z,ende BYTE
DIM position WORD

DIM sb0 BYTE   'Eigentlich mu jedes Byte des Buffers vor jedem Abspielen
DIM sb1 BYTE   '  auf 128 gesetzt werden.
DIM sb2 BYTE
DIM sb3 BYTE
DIM stemp BYTE

DEFINE laenge 56963

' ---------------------------------------------------------------------------
' Soundausgabe auf der Open-Maxi an DA[1]. Die folgenden Subroutinen sollten
' sinnvollerweise in einer INCLUDE-Datei untergebracht werden.
' ---------------------------------------------------------------------------

' Das Schlsselwort INLASM kennzeichnet die folgende PROCEDURE als
' Assemblerroutine ("IAR"). Dadurch, und da sie unmittelbar am Anfang
' des Programms steht, wird sie zu einer IIAR, also zu einer Assembler-
' Interruptroutine, die bei einem aktivierten User-Interrupt gestartet wird.

PROC iiar_tpm2ofl INLASM
! cbeqa #iTPM2OFL,is_tpm2ofl 'sind wir im passenden Interrupt?
! jmp iiar_tpm2ofl_next+2    'wenn nicht: zur naechsten IIAR springen

#is_tpm2ofl
! brset 0,ende,is_tpm2ofl_rts
! mov sb1,sb0                'Platz im Soundbuffer schaffen
! mov sb2,sb1
! mov sb3,sb2
! ldhx position
! lda soundtab,x
! sta sb3                    'aktuelles Sample in Buffer einfgen
! clr stemp                  '  (jedes Sample 4-mal)
! clrh
! ldx #3
#is_tpm2ofl_1
! add sb0-1,x                'fr Mittelwert die vier Samples addieren,
! bcc is_tpm2ofl_2           '  Ergebnis in stemp:a
! inc stemp                  '  (grtmglicher Wert: 4*255=1020)
#is_tpm2ofl_2
! dbnzx is_tpm2ofl_1
! lsr stemp                  'Ergebnis durch 2 teilen,
! rora                       '  ist jetzt maximal 510, also 9 Bit
! inc z
! brclr 2,z,is_tpm2ofl_setv  'bei jedem vierten Sample...
! clr z
! ldhx position              'Position in den Sampledaten erhhen
! aix #1
! sthx position
! cmphx #laenge              'ist das Ende der Sampledaten erreicht?
! bne is_tpm2ofl_setv
! mov #1,ende
#is_tpm2ofl_setv
! mov stemp,TPM2C0VH         'den berechneten Mittelwert ins "Timer 2
! sta TPM2C0VL               '  Channel 0 Value Register" schreiben,
#is_tpm2ofl_rts              '  um die Impulsbreite der PWM zu setzen
! clc
! rts                        'Rcksprung ins Betriebssystem

END PROC

' Das TPM2-Modul fr Soundausgabe konfigurieren

' DA[1] liegt an TPM2CH0
' DA[2] liegt an TPM2CH1

PROC inittpm2 INLASM
! mov #%01001000,TPM2SC      'TPM clock input = 20E6/1 = 20 MHz & Timer Overflow Interrupt Enable
! ldhx #623                  'Timer 2 Counter Modulo Registers (TPMxMODH:TPMxMODL)
! sthx TPM2MODH              'von 0 bis inklusive 623 zhlen (also Modulo 624) - (kein 100% duty cycle)
! lda #%00101000             'PWM-Frequenz: 20E6/1/624 = 32.051 kHz
! sta TPM2C0SC               'Timer 2 Channel 0 Status and Control Register \ Edge-aligned PWM
! sta TPM2C1SC               'Timer 2 Channel 1 Status and Control Register /  High-true pulses
! rts
END PROC

#iiar_tpm2ofl_next

' ---------------------------------------------------------------------------
' Ende der IIAR. Hier koennen weitere IIARs folgen.
' ---------------------------------------------------------------------------

#main          'durch die IIAR am Programmbeginn wird ein "GOTO main" erzeugt

' ---------------------------------------------------------------------------
' Das Hauptprogramm
' ---------------------------------------------------------------------------

position=1
ende=0
z=0
UIRTpm2Ofl=ON  'wir nutzen den "Timer/PWM-Module 2 Overflow"-Interrupt
inittpm2

PrintSpc=ON

WHILE TRUE
  PRINT "Spiele Sound an DA[1]";position;ende
  PAUSE 10
  IF ende THEN
    PRINT "Beginne von neuem"
    PAUSE 75
    position=1
    ende=0
    z=0
  END IF
WEND

END

' ---------------------------------------------------------------------------
' Sounddaten im Format "8-bit Unsigned", 8 kHz Sample-Frequenz
' ---------------------------------------------------------------------------

' Arnold Schwarzenegger als "Terminator".

TABLE soundtab BYTE "sound1.tab"

